using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;

namespace AsynchCalcPi2005
{
    class AsyncCalcPiForm5_Thread4_Async2_Safe : AsyncCalcPiForm4_Thread3_Async
    {
        public AsyncCalcPiForm5_Thread4_Async2_Safe()
            : base()
        {
            
        }

        protected virtual bool ShowProgress(string pi, int totalDigits, int digitsSoFar)
        {
            try
            {
                Debug.Assert(this.InvokeRequired == false);

                this.resultTextBox.Text = pi;
                this.calcToolStripProgressBar.Maximum = totalDigits;
                this.calcToolStripProgressBar.Value = digitsSoFar;

                if (digitsSoFar == totalDigits)
                {
                    this.calcToolStripStatusLabel.Text = "Ready";
                    this.calcToolStripProgressBar.Visible = false;
                }
                return true;
            }
            catch (NullReferenceException)
            {
                return false;
            }
        }

        protected delegate bool ShowProgressDelegate(
            string pi, int totalDigits, int digitsSoFar);

        override protected void CalcPi(int digits)
        {
            StringBuilder pi = new StringBuilder("3", digits + 2);

            ShowProgressDelegate showProgress =
                new ShowProgressDelegate(ShowProgress);
            
            
            this.Invoke(
                showProgress,
                new object[] {pi.ToString(), digits, 0});

            if (digits > 0)
            {
                pi.Append(".");

                for (int i = 0; i < digits; i += 9)
                {
                    int nineDigits = NineDigitsOfPi.StartingAt(i + 1);
                    int digitCount = Math.Min(digits - i, 9);
                    string ds = string.Format("{0:D9}", nineDigits);
                    pi.Append(ds.Substring(0, digitCount));

                    try
                    {
                        if (!(bool)this.Invoke(
                            showProgress,
                            new object[] { pi.ToString(), digits, i + digitCount }))
                        {
                            //form was closed, but operation is still running
                            Console.WriteLine("Still going after null pointer...");
                        }
                    }
                    catch (ObjectDisposedException)
                    {
                        //form was closed, but operation is still running
                        Console.WriteLine("Still going after dispose...");
                    }
                    catch (InvalidOperationException)
                    {
                        //form was closed, but operation is still running
                        Console.WriteLine("Still going after invalid operation...");
                    }
                }

            }
        }

        protected override void EndCalcPi(IAsyncResult result)
        {
            try
            {
                CalcPiDelegate calcPi = (CalcPiDelegate)result.AsyncState;
                calcPi.EndInvoke(result);
            }
            catch (Exception ex)
            {
                ShowProgressDelegate showProgress =
                new ShowProgressDelegate(ShowProgress);

                this.Invoke(
                    showProgress,
                    new object[] { ex.Message, 0, 0 });
            }
        }
    }
}
